@@ -8,12 +8,10 @@
 #include "hdy-combo-row.h"
 
 #include <glib/gi18n-lib.h>
-#include "hdy-list-box.h"
-#include "hdy-style-private.h"
 
 /**
  * SECTION:hdy-combo-row
- * @short_description: A #GtkListBox row used to choose from a list of items
+ * @short_description: A #GtkListBox row used to choose from a list of items.
  * @Title: HdyComboRow
  *
  * The #HdyComboRow widget allows the user to choose from a list of valid
@@ -24,6 +22,19 @@
  * is specified in the form of a #GListModel, and the display of the choices can
  * be adapted to the data in the model via widget creation functions.
  *
+ * #HdyComboRow is #GtkListBoxRow:activatable if a model is set.
+ *
+ * # CSS nodes
+ *
+ * #HdyComboRow has a main CSS node with name row.
+ *
+ * Its popover has the node name popover with the .combo style class, it
+ * contains a #GtkScrolledWindow, which in turn contains a #GtkListBox, both are
+ * accessible via their regular nodes.
+ *
+ * A checkmark of node and style class image.checkmark in the popover denotes
+ * the current item.
+ *
  * Since: 0.0.6
  */
 
@@ -31,6 +42,13 @@
  * This was mostly inspired by code from the display panel from GNOME Settings.
  */
 
+typedef struct
+{
+  HdyComboRowGetNameFunc func;
+  gpointer func_data;
+  GDestroyNotify func_data_destroy;
+} HdyComboRowGetName;
+
 typedef struct
 {
   GtkBox *current;
@@ -38,11 +56,18 @@ typedef struct
   GtkListBox *list;
   GtkPopover *popover;
   gint selected_index;
+  gboolean use_subtitle;
+  HdyComboRowGetName *get_name;
 
   GListModel *bound_model;
   GtkListBoxCreateWidgetFunc create_list_widget_func;
   GtkListBoxCreateWidgetFunc create_current_widget_func;
   gpointer create_widget_func_data;
+  GDestroyNotify create_widget_func_data_free_func;
+  /* This is owned by create_widget_func_data, which is ultimately owned by the
+   * list box, and hence should not be destroyed manually.
+   */
+  HdyComboRowGetName *get_name_internal;
 } HdyComboRowPrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE (HdyComboRow, hdy_combo_row, HDY_TYPE_ACTION_ROW)
@@ -50,32 +75,25 @@ G_DEFINE_TYPE_WITH_PRIVATE (HdyComboRow, hdy_combo_row, HDY_TYPE_ACTION_ROW)
 enum {
   PROP_0,
   PROP_SELECTED_INDEX,
+  PROP_USE_SUBTITLE,
   LAST_PROP,
 };
 
 static GParamSpec *props[LAST_PROP];
 
-typedef struct
-{
-  HdyComboRowGetNameFunc get_name_func;
-  gpointer get_name_func_data;
-  GDestroyNotify get_name_func_data_destroy;
-} HdyComboRowCreateLabelData;
-
 static GtkWidget *
 create_list_label (gpointer item,
                    gpointer user_data)
 {
-  HdyComboRowCreateLabelData *data = user_data;
-  g_autofree gchar *name = data->get_name_func (item, data->get_name_func_data);
+  HdyComboRowGetName *get_name = (HdyComboRowGetName *) user_data;
+  g_autofree gchar *name = get_name->func (item, get_name->func_data);
 
   return g_object_new (GTK_TYPE_LABEL,
                        "ellipsize", PANGO_ELLIPSIZE_END,
                        "label", name,
-                       "margin", 12,
                        "max-width-chars", 20,
+                       "valign", GTK_ALIGN_CENTER,
                        "visible", TRUE,
-                       "width-chars", 20,
                        "xalign", 0.0,
                         NULL);
 }
@@ -84,11 +102,11 @@ static GtkWidget *
 create_current_label (gpointer item,
                       gpointer user_data)
 {
-  HdyComboRowCreateLabelData *data = user_data;
+  HdyComboRowGetName *get_name = (HdyComboRowGetName *) user_data;
   g_autofree gchar *name = NULL;
 
-  if (data->get_name_func)
-    name = data->get_name_func (item, data->get_name_func_data);
+  if (get_name->func)
+    name = get_name->func (item, get_name->func_data);
 
   return g_object_new (GTK_TYPE_LABEL,
                        "ellipsize", PANGO_ELLIPSIZE_END,
@@ -101,19 +119,58 @@ create_current_label (gpointer item,
 }
 
 static void
-create_label_data_free (gpointer user_data)
+create_list_widget_data_free (gpointer user_data)
 {
-  HdyComboRowCreateLabelData *data = user_data;
+  HdyComboRow *self = HDY_COMBO_ROW (user_data);
+  HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
-  if (data == NULL)
+  if (priv->create_widget_func_data_free_func)
+    priv->create_widget_func_data_free_func (priv->create_widget_func_data);
+}
+
+static GtkWidget *
+create_list_widget (gpointer item,
+                    gpointer user_data)
+{
+  HdyComboRow *self = HDY_COMBO_ROW (user_data);
+  HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
+  GtkWidget *checkmark = g_object_new (GTK_TYPE_IMAGE,
+                                       "halign", GTK_ALIGN_START,
+                                       "icon-name", "object-select-symbolic",
+                                       "opacity", 0.0,
+                                       "valign", GTK_ALIGN_CENTER,
+                                       "visible", TRUE,
+                                       NULL);
+  GtkWidget *box = g_object_new (GTK_TYPE_BOX,
+                                 "child", priv->create_list_widget_func (item, priv->create_widget_func_data),
+                                 "child", checkmark,
+                                 "halign", GTK_ALIGN_START,
+                                 "spacing", 6,
+                                 "valign", GTK_ALIGN_CENTER,
+                                 "visible", TRUE,
+                                 NULL);
+  GtkStyleContext *checkmark_context = gtk_widget_get_style_context (checkmark);
+
+  gtk_style_context_add_class (checkmark_context, "checkmark");
+
+  g_object_set_data (G_OBJECT (box), "checkmark", checkmark);
+
+  return box;
+}
+
+static void
+get_name_free (HdyComboRowGetName *get_name)
+{
+  if (get_name == NULL)
     return;
 
-  if (data->get_name_func_data_destroy)
-    data->get_name_func_data_destroy (data->get_name_func_data);
-  data->get_name_func_data = NULL;
-  data->get_name_func_data_destroy = NULL;
+  if (get_name->func_data_destroy)
+    get_name->func_data_destroy (get_name->func_data);
+  get_name->func = NULL;
+  get_name->func_data = NULL;
+  get_name->func_data_destroy = NULL;
 
-  g_free (data);
+  g_free (get_name);
 }
 
 static void
@@ -121,24 +178,51 @@ update (HdyComboRow *self)
 {
   HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
   g_autoptr(GObject) item = NULL;
+  g_autofree gchar *name = NULL;
   GtkWidget *widget;
+  guint n_items = priv->bound_model ? g_list_model_get_n_items (priv->bound_model) : 0;
 
+  gtk_widget_set_visible (GTK_WIDGET (priv->current), !priv->use_subtitle);
   gtk_container_foreach (GTK_CONTAINER (priv->current), (GtkCallback) gtk_widget_destroy, NULL);
 
-  if (priv->bound_model == NULL || g_list_model_get_n_items (priv->bound_model) == 0) {
-    gtk_widget_set_sensitive (GTK_WIDGET (self), FALSE);
+  gtk_widget_set_sensitive (GTK_WIDGET (self), n_items > 0);
+  gtk_widget_set_visible (GTK_WIDGET (priv->image), n_items > 1);
+  gtk_list_box_row_set_activatable (GTK_LIST_BOX_ROW (self), n_items > 1);
+
+  if (n_items == 0) {
     g_assert (priv->selected_index == -1);
 
     return;
   }
 
-  g_assert (priv->selected_index >= 0 && priv->selected_index <= g_list_model_get_n_items (priv->bound_model));
+  g_assert (priv->selected_index >= 0 && priv->selected_index <= n_items);
 
-  gtk_widget_set_sensitive (GTK_WIDGET (self), TRUE);
+  {
+    g_autoptr (GList) rows = gtk_container_get_children (GTK_CONTAINER (priv->list));
+    GList *l;
+    int i = 0;
+
+    for (l = rows; l; l = l->next) {
+      GtkWidget *row = GTK_WIDGET (l->data);
+      GtkWidget *box = gtk_bin_get_child (GTK_BIN (row));
+
+      gtk_widget_set_opacity (GTK_WIDGET (g_object_get_data (G_OBJECT (box), "checkmark")),
+                              (priv->selected_index == i++) ? 1 : 0);
+    }
+  }
 
   item = g_list_model_get_item (priv->bound_model, priv->selected_index);
-  widget = priv->create_current_widget_func (item, priv->create_widget_func_data);
-  gtk_container_add (GTK_CONTAINER (priv->current), widget);
+  if (priv->use_subtitle) {
+    if (priv->get_name != NULL && priv->get_name->func)
+      name = priv->get_name->func (item, priv->get_name->func_data);
+    else if (priv->get_name_internal != NULL && priv->get_name_internal->func)
+      name = priv->get_name_internal->func (item, priv->get_name_internal->func_data);
+    hdy_action_row_set_subtitle (HDY_ACTION_ROW (self), name);
+  }
+  else {
+    widget = priv->create_current_widget_func (item, priv->create_widget_func_data);
+    gtk_container_add (GTK_CONTAINER (priv->current), widget);
+  }
 }
 
 static void
@@ -153,8 +237,10 @@ bound_model_changed (GListModel *list,
   HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
   /* Selection is in front of insertion/removal point, nothing to do */
-  if (priv->selected_index > 0 && priv->selected_index < index)
+  if (priv->selected_index > 0 && priv->selected_index < index) {
+    update (self);
     return;
+  }
 
   if (priv->selected_index < index + removed) {
     /* The item selected item was removed (or none is selected) */
@@ -169,6 +255,8 @@ bound_model_changed (GListModel *list,
     new_idx = 0;
 
   hdy_combo_row_set_selected_index (self, new_idx);
+
+  update (self);
 }
 
 static void
@@ -184,7 +272,8 @@ hdy_combo_row_activate (HdyActionRow *row)
   HdyComboRow *self = HDY_COMBO_ROW (row);
   HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
-  gtk_popover_popup (priv->popover);
+  if (gtk_widget_get_visible (GTK_WIDGET (priv->image)))
+    gtk_popover_popup (priv->popover);
 }
 
 static void
@@ -192,19 +281,21 @@ destroy_model (HdyComboRow *self)
 {
   HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
-  if (priv->bound_model) {
-    /* Disconnect the bound model *before* releasing it. */
-    g_signal_handlers_disconnect_by_func (priv->bound_model, bound_model_changed, self);
+  if (!priv->bound_model)
+    return;
 
-    /* Destroy the model and the user data. */
-    if (priv->list)
-      gtk_list_box_bind_model (priv->list, NULL, NULL, NULL, NULL);
+  /* Disconnect the bound model *before* releasing it. */
+  g_signal_handlers_disconnect_by_func (priv->bound_model, bound_model_changed, self);
 
-    priv->bound_model = NULL;
-    priv->create_list_widget_func = NULL;
-    priv->create_current_widget_func = NULL;
-    priv->create_widget_func_data = NULL;
-  }
+  /* Destroy the model and the user data. */
+  if (priv->list)
+    gtk_list_box_bind_model (priv->list, NULL, NULL, NULL, NULL);
+
+  priv->bound_model = NULL;
+  priv->create_list_widget_func = NULL;
+  priv->create_current_widget_func = NULL;
+  priv->create_widget_func_data = NULL;
+  priv->create_widget_func_data_free_func = NULL;
 }
 
 static void
@@ -219,6 +310,9 @@ hdy_combo_row_get_property (GObject    *object,
   case PROP_SELECTED_INDEX:
     g_value_set_int (value, hdy_combo_row_get_selected_index (self));
     break;
+  case PROP_USE_SUBTITLE:
+    g_value_set_boolean (value, hdy_combo_row_get_use_subtitle (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -236,6 +330,9 @@ hdy_combo_row_set_property (GObject      *object,
   case PROP_SELECTED_INDEX:
     hdy_combo_row_set_selected_index (self, g_value_get_int (value));
     break;
+  case PROP_USE_SUBTITLE:
+    hdy_combo_row_set_use_subtitle (self, g_value_get_boolean (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -245,8 +342,10 @@ static void
 hdy_combo_row_dispose (GObject *object)
 {
   HdyComboRow *self = HDY_COMBO_ROW (object);
+  HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
   destroy_model (self);
+  g_clear_pointer (&priv->get_name, get_name_free);
 
   G_OBJECT_CLASS (hdy_combo_row_parent_class)->dispose (object);
 }
@@ -321,6 +420,25 @@ hdy_combo_row_class_init (HdyComboRowClass *klass)
                         -1, G_MAXINT, -1,
                         G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * HdyComboRow:use-subtitle:
+   *
+   * %TRUE to set the current value as the subtitle.
+   *
+   * If you use a custom widget creation function, you will need to give the row
+   * a name conversion closure with hdy_combo_row_set_get_name_func().
+   *
+   * If %TRUE, you should not access HdyActionRow:subtitle.
+   *
+   * Since: 0.0.10
+   */
+  props[PROP_USE_SUBTITLE] =
+    g_param_spec_boolean ("use-subtitle",
+                          _("Use subtitle"),
+                          _("Set the current value as the subtitle"),
+                          FALSE,
+                          G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
   gtk_widget_class_set_template_from_resource (widget_class,
@@ -331,18 +449,6 @@ hdy_combo_row_class_init (HdyComboRowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, HdyComboRow, popover);
 }
 
-static void
-list_init (HdyComboRow *self)
-{
-  HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
-  g_autoptr (GtkCssProvider) provider = gtk_css_provider_new ();
-
-  gtk_css_provider_load_from_resource (provider, "/sm/puri/handy/style/hdy-combo-row-list.css");
-  gtk_style_context_add_provider (gtk_widget_get_style_context (GTK_WIDGET (priv->list)),
-                                  GTK_STYLE_PROVIDER (provider),
-                                  HDY_STYLE_PROVIDER_PRIORITY);
-}
-
 static void
 hdy_combo_row_init (HdyComboRow *self)
 {
@@ -350,12 +456,9 @@ hdy_combo_row_init (HdyComboRow *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  list_init (self);
-
   priv->selected_index = -1;
 
-  gtk_list_box_set_header_func (priv->list, hdy_list_box_separator_header, NULL, NULL);
-  g_signal_connect_object (priv->list, "row-activated", G_CALLBACK (gtk_widget_hide),
+  g_signal_connect_object (priv->list, "row-activated", G_CALLBACK (gtk_popover_popdown),
                            priv->popover, G_CONNECT_SWAPPED);
   g_signal_connect_object (priv->list, "row-activated", G_CALLBACK (row_activated_cb),
                            self, G_CONNECT_SWAPPED);
@@ -372,7 +475,7 @@ hdy_combo_row_init (HdyComboRow *self)
  *
  * Since: 0.0.6
  */
-HdyComboRow *
+GtkWidget *
 hdy_combo_row_new (void)
 {
   return g_object_new (HDY_TYPE_COMBO_ROW, NULL);
@@ -408,8 +511,8 @@ hdy_combo_row_get_model (HdyComboRow *self)
  *   widgets for items to display in the list, or %NULL in case you also passed
  *   %NULL as @model
  * @create_current_widget_func: (nullable) (scope call): a function that creates
- *   widgets for items to display as the seleted item, or %NULL in case you also
- *   passed %NULL as @model
+ *   widgets for items to display as the selected item, or %NULL in case you
+ *   also passed %NULL as @model
  * @user_data: user data passed to @create_list_widget_func and
  *   @create_current_widget_func
  * @user_data_free_func: function for freeing @user_data
@@ -453,19 +556,20 @@ hdy_combo_row_bind_model (HdyComboRow                *self,
     return;
   }
 
-  gtk_list_box_bind_model (priv->list, model, create_list_widget_func, user_data, user_data_free_func);
-
   /* We don't need take a reference as the list box holds one for us. */
   priv->bound_model = model;
   priv->create_list_widget_func = create_list_widget_func;
   priv->create_current_widget_func = create_current_widget_func;
   priv->create_widget_func_data = user_data;
+  priv->create_widget_func_data_free_func = user_data_free_func;
 
   g_signal_connect (priv->bound_model, "items-changed", G_CALLBACK (bound_model_changed), self);
 
   if (g_list_model_get_n_items (priv->bound_model) > 0)
     priv->selected_index = 0;
 
+  gtk_list_box_bind_model (priv->list, model, create_list_widget, self, create_list_widget_data_free);
+
   update (self);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SELECTED_INDEX]);
@@ -488,8 +592,8 @@ hdy_combo_row_bind_model (HdyComboRow                *self,
  * items from @model. @self is updated whenever @model changes. If @model is
  * %NULL, @self is left empty.
  *
- * This is more conventient to use than hdy_combo_row_bind_model() if you want
- * to represent items of the model with names.
+ * This is more convenient to use than hdy_combo_row_bind_model() if you want to
+ * represent items of the model with names.
  *
  * Since: 0.0.6
  */
@@ -500,18 +604,18 @@ hdy_combo_row_bind_name_model (HdyComboRow            *self,
                                gpointer                user_data,
                                GDestroyNotify          user_data_free_func)
 {
-  HdyComboRowCreateLabelData *data;
+  HdyComboRowPrivate *priv = hdy_combo_row_get_instance_private (self);
 
   g_return_if_fail (HDY_IS_COMBO_ROW (self));
   g_return_if_fail (model == NULL || G_IS_LIST_MODEL (model));
   g_return_if_fail (model == NULL || get_name_func != NULL);
 
-  data = g_new0 (HdyComboRowCreateLabelData, 1);
-  data->get_name_func = get_name_func;
-  data->get_name_func_data = user_data;
-  data->get_name_func_data_destroy = user_data_free_func;
+  priv->get_name_internal = g_new0 (HdyComboRowGetName, 1);
+  priv->get_name_internal->func = get_name_func;
+  priv->get_name_internal->func_data = user_data;
+  priv->get_name_internal->func_data_destroy = user_data_free_func;
 
-  hdy_combo_row_bind_model (self, model, create_list_label, create_current_label, data, create_label_data_free);
+  hdy_combo_row_bind_model (self, model, create_list_label, create_current_label, priv->get_name_internal, (GDestroyNotify) get_name_free);
 }
 
 /**
@@ -532,7 +636,7 @@ hdy_combo_row_bind_name_model (HdyComboRow            *self,
  * items from @model. @self is updated whenever @model changes. If @model is
  * %NULL, @self is left empty.
  *
- * This is more conventient to use than hdy_combo_row_bind_name_model() if you
+ * This is more convenient to use than hdy_combo_row_bind_name_model() if you
  * want to represent values of an enumeration with names.
  *
  * See hdy_enum_value_row_name().
@@ -619,6 +723,93 @@ hdy_combo_row_set_selected_index (HdyComboRow *self,
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SELECTED_INDEX]);
 }
 
+/**
+ * hdy_combo_row_get_use_subtitle:
+ * @self: a #GtkListBoxRow
+ *
+ * Gets whether the current value of @self should be displayed as its subtitle.
+ *
+ * Returns: whether the current value of @self should be displayed as its subtitle
+ *
+ * Since: 0.0.10
+ */
+gboolean
+hdy_combo_row_get_use_subtitle (HdyComboRow *self)
+{
+  HdyComboRowPrivate *priv;
+
+  g_return_val_if_fail (HDY_IS_COMBO_ROW (self), FALSE);
+
+  priv = hdy_combo_row_get_instance_private (self);
+
+  return priv->use_subtitle;
+}
+
+/**
+ * hdy_combo_row_set_use_subtitle:
+ * @self: a #HdyComboRow
+ * @use_subtitle: %TRUE to set the current value as the subtitle
+ *
+ * Sets whether the current value of @self should be displayed as its subtitle.
+ *
+ * If %TRUE, you should not access HdyActionRow:subtitle.
+ *
+ * Since: 0.0.10
+ */
+void
+hdy_combo_row_set_use_subtitle (HdyComboRow *self,
+                                gboolean     use_subtitle)
+{
+  HdyComboRowPrivate *priv;
+
+  g_return_if_fail (HDY_IS_COMBO_ROW (self));
+
+  priv = hdy_combo_row_get_instance_private (self);
+
+  use_subtitle = !!use_subtitle;
+
+  if (priv->use_subtitle == use_subtitle)
+    return;
+
+  priv->use_subtitle = use_subtitle;
+  update (self);
+  if (!use_subtitle)
+    hdy_action_row_set_subtitle (HDY_ACTION_ROW (self), NULL);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_USE_SUBTITLE]);
+}
+
+/**
+ * hdy_combo_row_set_get_name_func:
+ * @self: a #HdyComboRow
+ * @get_name_func: (nullable): a function that creates names for items, or %NULL
+ *   in case you also passed %NULL as @model
+ * @user_data: user data passed to @get_name_func
+ * @user_data_free_func: function for freeing @user_data
+ *
+ * Sets a closure to convert items into names. See HdyComboRow:use-subtitle.
+ *
+ * Since: 0.0.10
+ */
+void
+hdy_combo_row_set_get_name_func (HdyComboRow            *self,
+                                 HdyComboRowGetNameFunc  get_name_func,
+                                 gpointer                user_data,
+                                 GDestroyNotify          user_data_free_func)
+{
+  HdyComboRowPrivate *priv;
+
+  g_return_if_fail (HDY_IS_COMBO_ROW (self));
+
+  priv = hdy_combo_row_get_instance_private (self);
+
+  get_name_free (priv->get_name);
+  priv->get_name = g_new0 (HdyComboRowGetName, 1);
+  priv->get_name->func = get_name_func;
+  priv->get_name->func_data = user_data;
+  priv->get_name->func_data_destroy = user_data_free_func;
+}
+
 /**
  * hdy_enum_value_row_name:
  * @value: the value from the enum from which to get a name
