@@ -25,8 +25,10 @@
 #include "hdy-header-bar.h"
 
 #include "hdy-animation-private.h"
-#include "hdy-dialog.h"
+#include "hdy-cairo-private.h"
+#include "hdy-css-private.h"
 #include "hdy-enums.h"
+#include "hdy-window-handle-controller-private.h"
 #include "gtkprogresstrackerprivate.h"
 #include "gtk-window-private.h"
 
@@ -34,7 +36,7 @@
  * SECTION:hdy-header-bar
  * @short_description: A box with a centered child.
  * @Title: HdyHeaderBar
- * @See_also: #GtkHeaderBar, #HdyTitleBar, #HdyViewSwitcher, #HdyDialog
+ * @See_also: #GtkHeaderBar, #HdyApplicationWindow, #HdyTitleBar, #HdyViewSwitcher, #HdyWindow
  *
  * HdyHeaderBar is similar to #GtkHeaderBar but is designed to fix some of its
  * shortcomings for adaptive applications.
@@ -43,9 +45,19 @@
  * hence allowing it to fill up the whole height, which is e.g. needed for
  * #HdyViewSwitcher.
  *
- * When used in a #HdyDialog, HdyHeaderBar will replace its window decorations
- * by a back button allowing to close it. It doesn't have to be its direct child
- * and you can use any complex contraption you like as the dialog's titlebar.
+ * When used in a mobile dialog, HdyHeaderBar will replace its window
+ * decorations by a back button allowing to close it. It doesn't have to be its
+ * direct child and you can use any complex contraption you like as the dialog's
+ * titlebar.
+ *
+ * #HdyHeaderBar can be used in window's content area rather than titlebar, and
+ * will still be draggable and will handle right click, middle click and double
+ * click as expected from a titlebar. This is particularly useful with
+ * #HdyWindow or #HdyApplicationWindow.
+ *
+ * # CSS nodes
+ *
+ * #HdyHeaderBar has a single CSS node with name headerbar.
  */
 
 /**
@@ -57,7 +69,7 @@
 #define DEFAULT_SPACING 6
 #define MIN_TITLE_CHARS 5
 
-#define MOBILE_WINDOW_WIDTH  400
+#define MOBILE_WINDOW_WIDTH  480
 #define MOBILE_WINDOW_HEIGHT 800
 
 typedef struct {
@@ -97,6 +109,8 @@ typedef struct {
   gboolean is_mobile_window;
 
   gulong window_size_allocated_id;
+
+  HdyWindowHandleController *controller;
 } HdyHeaderBarPrivate;
 
 typedef struct _Child Child;
@@ -314,12 +328,10 @@ hdy_header_bar_update_window_icon (HdyHeaderBar *self,
     pixbuf = hdy_gtk_window_get_icon_for_size (window, scale * 20);
 
   if (pixbuf) {
-    cairo_surface_t *surface;
-
-    surface = gdk_cairo_surface_create_from_pixbuf (pixbuf, scale, gtk_widget_get_window (priv->titlebar_icon));
+    g_autoptr (cairo_surface_t) surface =
+      gdk_cairo_surface_create_from_pixbuf (pixbuf, scale, gtk_widget_get_window (priv->titlebar_icon));
 
     gtk_image_set_from_surface (GTK_IMAGE (priv->titlebar_icon), surface);
-    cairo_surface_destroy (surface);
     g_object_unref (pixbuf);
     gtk_widget_show (priv->titlebar_icon);
 
@@ -602,7 +614,7 @@ compute_is_mobile_window (GtkWindow *window)
   gtk_window_get_size (window, &window_width, &window_height);
 
   if (window_width <= MOBILE_WINDOW_WIDTH &&
-      window_height <= MOBILE_WINDOW_HEIGHT)
+      gtk_window_is_maximized (window))
     return TRUE;
 
   /* Mobile landscape mode. */
@@ -716,7 +728,7 @@ hdy_header_bar_get_size (GtkWidget      *widget,
   HdyHeaderBar *self = HDY_HEADER_BAR (widget);
   HdyHeaderBarPrivate *priv = hdy_header_bar_get_instance_private (self);
   GList *l;
-  gint n_start_children = 0, n_end_children = 0, n_center_children = 0;
+  gint n_start_children = 0, n_end_children = 0;
   gint start_min = 0, start_nat = 0;
   gint end_min = 0, end_nat = 0;
   gint center_min = 0, center_nat = 0;
@@ -738,15 +750,10 @@ hdy_header_bar_get_size (GtkWidget      *widget,
       add_child_size (priv->label_box, orientation, &center_min, &center_nat);
     else
       add_child_size (priv->label_sizing_box, orientation, &center_min, &center_nat);
-
-    if (gtk_widget_get_visible (priv->label_sizing_box))
-      n_center_children += 1;
   }
 
-  if (priv->custom_title != NULL) {
-    if (add_child_size (priv->custom_title, orientation, &center_min, &center_nat))
-      n_center_children += 1;
-  }
+  if (priv->custom_title != NULL)
+    add_child_size (priv->custom_title, orientation, &center_min, &center_nat);
 
   if (priv->titlebar_start_box != NULL) {
     if (add_child_size (priv->titlebar_start_box, orientation, &start_min, &start_nat))
@@ -773,12 +780,12 @@ hdy_header_bar_get_size (GtkWidget      *widget,
       strict_centering_t = priv->centering_policy == HDY_CENTERING_POLICY_STRICT ? 1.0 : 0.0;
 
     *minimum = center_min + n_start_children * priv->spacing +
-               hdy_lerp (2 * MAX (start_min_spaced, end_min_spaced),
-                         start_min_spaced + end_min_spaced,
+               hdy_lerp (start_min_spaced + end_min_spaced,
+                         2 * MAX (start_min_spaced, end_min_spaced),
                          strict_centering_t);
     *natural = center_nat + n_start_children * priv->spacing +
-               hdy_lerp (2 * MAX (start_nat_spaced, end_nat_spaced),
-                         start_nat_spaced + end_nat_spaced,
+               hdy_lerp (start_nat_spaced + end_nat_spaced,
+                         2 * MAX (start_nat_spaced, end_nat_spaced),
                          strict_centering_t);
   } else {
     *minimum = MAX (MAX (start_min, end_min), center_min);
@@ -868,7 +875,6 @@ hdy_header_bar_compute_size_for_opposing_orientation (GtkWidget *widget,
   gint computed_natural = 0;
   GtkRequestedSize *sizes;
   GtkPackType packing;
-  gint size = 0;
   gint i;
   gint child_size;
   gint child_minimum;
@@ -891,14 +897,13 @@ hdy_header_bar_compute_size_for_opposing_orientation (GtkWidget *widget,
                                       &sizes[i].minimum_size,
                                       &sizes[i].natural_size);
 
-      size -= sizes[i].minimum_size;
       sizes[i].data = child;
       i += 1;
     }
   }
 
   /* Bring children up to size first */
-  size = gtk_distribute_natural_allocation (MAX (0, avail_size), nvis_children, sizes);
+  gtk_distribute_natural_allocation (MAX (0, avail_size), nvis_children, sizes);
 
   /* Allocate child positions. */
   for (packing = GTK_PACK_START; packing <= GTK_PACK_END; ++packing) {
@@ -926,7 +931,6 @@ hdy_header_bar_compute_size_for_opposing_orientation (GtkWidget *widget,
       computed_minimum = MAX (computed_minimum, child_minimum);
       computed_natural = MAX (computed_natural, child_natural);
     }
-    i += 1;
   }
 
   center_min = center_nat = 0;
@@ -968,36 +972,22 @@ hdy_header_bar_measure (GtkWidget      *widget,
                         gint           *minimum_baseline,
                         gint           *natural_baseline)
 {
-  GtkStyleContext *style_context;
-  GtkStateFlags state_flags;
-  GtkBorder border, margin, padding;
+  gint css_width, css_height;
+
+  gtk_style_context_get (gtk_widget_get_style_context (widget),
+                         gtk_widget_get_state_flags (widget),
+                         "min-width", &css_width,
+                         "min-height", &css_height,
+                         NULL);
 
   if (for_size < 0)
     hdy_header_bar_get_size (widget, orientation, minimum, natural);
   else if (orientation == GTK_ORIENTATION_HORIZONTAL)
-    hdy_header_bar_compute_size_for_orientation (widget, for_size, minimum, natural);
+    hdy_header_bar_compute_size_for_orientation (widget, MAX (for_size, css_height), minimum, natural);
   else
-    hdy_header_bar_compute_size_for_opposing_orientation (widget, for_size, minimum, natural);
+    hdy_header_bar_compute_size_for_opposing_orientation (widget, MAX (for_size, css_width), minimum, natural);
 
-  /* Manually apply the border, the padding and the margin as we can't use the
-   * private GtkGagdet.
-   */
-  style_context = gtk_widget_get_style_context (widget);
-  state_flags = gtk_widget_get_state_flags (widget);
-  gtk_style_context_get_border (style_context, state_flags, &border);
-  gtk_style_context_get_margin (style_context, state_flags, &margin);
-  gtk_style_context_get_padding (style_context, state_flags, &padding);
-  if (orientation == GTK_ORIENTATION_VERTICAL) {
-    *minimum += border.top + margin.top + padding.top +
-                border.bottom + margin.bottom + padding.bottom;
-    *natural += border.top + margin.top + padding.top +
-                border.bottom + margin.bottom + padding.bottom;
-  } else {
-    *minimum += border.left + margin.left + padding.left +
-                border.right + margin.right + padding.right;
-    *natural += border.left + margin.left + padding.left +
-                border.right + margin.right + padding.right;
-  }
+  hdy_css_measure (widget, orientation, minimum, natural);
 }
 
 static void
@@ -1428,36 +1418,35 @@ hdy_header_bar_size_allocate (GtkWidget     *widget,
   HdyHeaderBarPrivate *priv = hdy_header_bar_get_instance_private (self);
   GtkAllocation *allocations;
   GtkAllocation title_allocation;
+  GtkAllocation clip;
   gint nvis_children;
   GList *l;
   gint i;
   Child *child;
   GtkAllocation child_allocation;
   GtkTextDirection direction;
-  GtkStyleContext *style_context;
-  GtkStateFlags state_flags;
-  GtkBorder border, margin, padding;
   GtkWidget *decoration_box[2] = { priv->titlebar_start_box, priv->titlebar_end_box };
   gint decoration_width[2] = { 0 };
 
+  gtk_render_background_get_clip (gtk_widget_get_style_context (widget),
+                                  allocation->x,
+                                  allocation->y,
+                                  allocation->width,
+                                  allocation->height,
+                                  &clip);
+
   gtk_widget_set_allocation (widget, allocation);
 
-  /* Manually apply the border, the padding and the margin as we can't use the
-   * private GtkGagdet.
-   */
-  style_context = gtk_widget_get_style_context (widget);
-  state_flags = gtk_widget_get_state_flags (widget);
-  gtk_style_context_get_border (style_context, state_flags, &border);
-  gtk_style_context_get_margin (style_context, state_flags, &margin);
-  gtk_style_context_get_padding (style_context, state_flags, &padding);
-  allocation->width -= border.left + border.right +
-                       margin.left + margin.right +
-                       padding.left + padding.right;
-  allocation->height -= border.top + border.bottom +
-                        margin.top + margin.bottom +
-                        padding.top + padding.bottom;
-  allocation->x += border.left + margin.left + padding.left;
-  allocation->y += border.top + margin.top + padding.top;
+  if (gtk_widget_get_realized (widget))
+    gdk_window_move_resize (gtk_widget_get_window (widget),
+                            allocation->x,
+                            allocation->y,
+                            allocation->width,
+                            allocation->height);
+
+  allocation->x = 0;
+  allocation->y = 0;
+  hdy_css_size_allocate (widget, allocation);
 
   direction = gtk_widget_get_direction (widget);
   nvis_children = count_visible_children (self);
@@ -1520,15 +1509,15 @@ hdy_header_bar_size_allocate (GtkWidget     *widget,
     get_strict_centering_allocations (self, allocation, &strict_allocations, &strict_title_allocation, decoration_width);
 
     for (i = 0; i < nvis_children; i++) {
-      allocations[i].x = hdy_lerp (strict_allocations[i].x, allocations[i].x, strict_centering_t);
-      allocations[i].y = hdy_lerp (strict_allocations[i].y, allocations[i].y, strict_centering_t);
-      allocations[i].width = hdy_lerp (strict_allocations[i].width, allocations[i].width, strict_centering_t);
-      allocations[i].height = hdy_lerp (strict_allocations[i].height, allocations[i].height, strict_centering_t);
+      allocations[i].x = hdy_lerp (allocations[i].x, strict_allocations[i].x, strict_centering_t);
+      allocations[i].y = hdy_lerp (allocations[i].y, strict_allocations[i].y, strict_centering_t);
+      allocations[i].width = hdy_lerp (allocations[i].width, strict_allocations[i].width, strict_centering_t);
+      allocations[i].height = hdy_lerp (allocations[i].height, strict_allocations[i].height, strict_centering_t);
     }
-    title_allocation.x = hdy_lerp (strict_title_allocation.x, title_allocation.x, strict_centering_t);
-    title_allocation.y = hdy_lerp (strict_title_allocation.y, title_allocation.y, strict_centering_t);
-    title_allocation.width = hdy_lerp (strict_title_allocation.width, title_allocation.width, strict_centering_t);
-    title_allocation.height = hdy_lerp (strict_title_allocation.height, title_allocation.height, strict_centering_t);
+    title_allocation.x = hdy_lerp (title_allocation.x, strict_title_allocation.x, strict_centering_t);
+    title_allocation.y = hdy_lerp (title_allocation.y, strict_title_allocation.y, strict_centering_t);
+    title_allocation.width = hdy_lerp (title_allocation.width, strict_title_allocation.width, strict_centering_t);
+    title_allocation.height = hdy_lerp (title_allocation.height, strict_title_allocation.height, strict_centering_t);
   }
 
   /* Allocate the children on both sides of the title. */
@@ -1547,6 +1536,8 @@ hdy_header_bar_size_allocate (GtkWidget     *widget,
     gtk_widget_size_allocate (priv->custom_title, &title_allocation);
   else if (priv->label_box != NULL)
     gtk_widget_size_allocate (priv->label_box, &title_allocation);
+
+  gtk_widget_set_clip (widget, &clip);
 }
 
 static void
@@ -1592,6 +1583,7 @@ hdy_header_bar_finalize (GObject *object)
   g_clear_pointer (&priv->title, g_free);
   g_clear_pointer (&priv->subtitle, g_free);
   g_clear_pointer (&priv->decoration_layout, g_free);
+  g_clear_object (&priv->controller);
 
   G_OBJECT_CLASS (hdy_header_bar_parent_class)->finalize (object);
 }
@@ -1941,7 +1933,7 @@ hdy_header_bar_draw (GtkWidget *widget,
 
   context = gtk_widget_get_style_context (widget);
   /* GtkWidget draws nothing by default so we have to render the background
-   * explicitely for HdyHederBar to render the typical titlebar background.
+   * explicitly for HdyHederBar to render the typical titlebar background.
    */
   gtk_render_background (context,
                          cr,
@@ -1962,8 +1954,10 @@ static void
 hdy_header_bar_realize (GtkWidget *widget)
 {
   GtkSettings *settings;
-
-  GTK_WIDGET_CLASS (hdy_header_bar_parent_class)->realize (widget);
+  GtkAllocation allocation;
+  GdkWindowAttr attributes;
+  gint attributes_mask;
+  GdkWindow *window;
 
   settings = gtk_widget_get_settings (widget);
   g_signal_connect_swapped (settings, "notify::gtk-shell-shows-app-menu",
@@ -1972,6 +1966,25 @@ hdy_header_bar_realize (GtkWidget *widget)
                             G_CALLBACK (hdy_header_bar_update_window_buttons), widget);
   update_is_mobile_window (HDY_HEADER_BAR (widget));
   hdy_header_bar_update_window_buttons (HDY_HEADER_BAR (widget));
+
+  gtk_widget_get_allocation (widget, &allocation);
+  gtk_widget_set_realized (widget, TRUE);
+
+  attributes.x = allocation.x;
+  attributes.y = allocation.y;
+  attributes.width = allocation.width;
+  attributes.height = allocation.height;
+  attributes.window_type = GDK_WINDOW_CHILD;
+  attributes.event_mask = gtk_widget_get_events (widget);
+  attributes.visual = gtk_widget_get_visual (widget);
+  attributes.wclass = GDK_INPUT_OUTPUT;
+  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL;
+
+  window = gdk_window_new (gtk_widget_get_parent_window (widget),
+                           &attributes,
+                           attributes_mask);
+  gtk_widget_set_window (widget, window);
+  gtk_widget_register_window (widget, window);
 }
 
 static void
@@ -2216,11 +2229,10 @@ static void
 hdy_header_bar_init (HdyHeaderBar *self)
 {
   HdyHeaderBarPrivate *priv;
+  GtkStyleContext *context;
 
   priv = hdy_header_bar_get_instance_private (self);
 
-  gtk_widget_set_has_window (GTK_WIDGET (self), FALSE);
-
   priv->title = NULL;
   priv->subtitle = NULL;
   priv->custom_title = NULL;
@@ -2234,6 +2246,11 @@ hdy_header_bar_init (HdyHeaderBar *self)
   init_sizing_box (self);
   construct_label_box (self);
 
+  priv->controller = hdy_window_handle_controller_new (GTK_WIDGET (self));
+
+  context = gtk_widget_get_style_context (GTK_WIDGET (self));
+  /* Ensure the widget has the titlebar style class. */
+  gtk_style_context_add_class (context, "titlebar");
 }
 
 static void
@@ -2308,7 +2325,7 @@ hdy_header_bar_pack_end (HdyHeaderBar *self,
 /**
  * hdy_header_bar_set_title:
  * @self: a #HdyHeaderBar
- * @title: (allow-none): a title, or %NULL
+ * @title: (nullable): a title, or %NULL
  *
  * Sets the title of the #HdyHeaderBar. The title should help a user
  * identify the current view. A good title should not include the
@@ -2362,10 +2379,10 @@ hdy_header_bar_get_title (HdyHeaderBar *self)
 /**
  * hdy_header_bar_set_subtitle:
  * @self: a #HdyHeaderBar
- * @subtitle: (allow-none): a subtitle, or %NULL
+ * @subtitle: (nullable): a subtitle, or %NULL
  *
  * Sets the subtitle of the #HdyHeaderBar. The title should give a user
- * an additional detail to help him identify the current view.
+ * an additional detail to help them identify the current view.
  *
  * Note that HdyHeaderBar by default reserves room for the subtitle,
  * even if none is currently set. If this is not desired, set the
@@ -2422,7 +2439,7 @@ hdy_header_bar_get_subtitle (HdyHeaderBar *self)
 /**
  * hdy_header_bar_set_custom_title:
  * @self: a #HdyHeaderBar
- * @title_widget: (allow-none): a custom widget to use for a title
+ * @title_widget: (nullable): a custom widget to use for a title
  *
  * Sets a custom title for the #HdyHeaderBar.
  *
@@ -2616,8 +2633,7 @@ hdy_header_bar_get_has_subtitle (HdyHeaderBar *self)
 /**
  * hdy_header_bar_set_decoration_layout:
  * @self: a #HdyHeaderBar
- * @layout: (allow-none): a decoration layout, or %NULL to
- *     unset the layout
+ * @layout: (nullable): a decoration layout, or %NULL to unset the layout
  *
  * Sets the decoration layout for this header bar, overriding
  * the #GtkSettings:gtk-decoration-layout setting.
@@ -2801,7 +2817,7 @@ hdy_header_bar_get_transition_running (HdyHeaderBar *self)
  * hdy_header_bar_get_interpolate_size:
  * @self: A #HdyHeaderBar
  *
- * Gets wether @self should interpolate its size on visible child change.
+ * Gets whether @self should interpolate its size on visible child change.
  *
  * See hdy_header_bar_set_interpolate_size().
  *
